import unittest
import os
import filecmp
from ariba import report_flag_expander

modules_dir = os.path.dirname(os.path.abspath(report_flag_expander.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestReportFlagExpander(unittest.TestCase):
    def test_run(self):
        '''test run'''
        infile = os.path.join(data_dir, 'report_flag_expander.run.in.tsv')
        expected = os.path.join(data_dir, 'report_flag_expander.run.out.tsv')
        tmp_out = 'tmp.report_flag_expander.out.tsv'
        expander = report_flag_expander.ReportFlagExpander(infile, tmp_out)
        expander.run()
        self.assertTrue(filecmp.cmp(expected, tmp_out, shallow=False))
        os.unlink(tmp_out)




import unittest
import os
from ariba import card_record

modules_dir = os.path.dirname(os.path.abspath(card_record.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestCardRecord(unittest.TestCase):
    def test_ARO_id(self):
        '''test _ARO_id'''
        d = {'spam': 'eggs'}
        self.assertEqual(None, card_record.CardRecord._ARO_id(d))
        d['ARO_id'] = '123'
        self.assertEqual('123', card_record.CardRecord._ARO_id(d))


    def test_ARO_accession(self):
        '''test _ARO_accession'''
        d = {'spam': 'eggs'}
        self.assertEqual(None, card_record.CardRecord._ARO_accession(d))
        d['ARO_accession'] = '321'
        self.assertEqual('321', card_record.CardRecord._ARO_accession(d))


    def test_ARO_name(self):
        '''test _ARO_name'''
        d = {'spam': 'eggs'}
        self.assertEqual(None, card_record.CardRecord._ARO_name(d))
        d['ARO_name'] = 'Dave Lister'
        self.assertEqual('Dave Lister', card_record.CardRecord._ARO_name(d))


    def test_ARO_description(self):
        '''test _ARO_description'''
        d = {'spam': 'eggs'}
        self.assertEqual(None, card_record.CardRecord._ARO_description(d))
        d['ARO_description'] = 'Technician, Third Class'
        self.assertEqual('Technician, Third Class', card_record.CardRecord._ARO_description(d))


    def test_ARO_name_to_fasta_name(self):
        '''test _ARO_name_to_fasta_name'''
        tests = [
            ('x', 'x'),
            ('abcD foo bar match at the start', 'abcD'),
            ('foo bar abcD match in the middle', 'abcD'),
            ('match at the end abcD', 'abcD'),
            ('use first three foo bar', 'use_first_three'),
            ('remove.any.dots first three', 'remove_any_dots_first_three')
        ]

        for aro_name, expected in tests:
            got = card_record.CardRecord._ARO_name_to_fasta_name(aro_name)
            self.assertEqual(expected, got)


    def test_dna_seqs_and_genbank_ids(self):
        '''test _dna_seqs_and_genbank_ids'''
        d = {'spam': 'eggs'}
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences'] = {}
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence'] = {}
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo'] = {}
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['dna_sequence'] = {}
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['dna_sequence']['sequence'] = 'ACGT'
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['dna_sequence']['accession'] = 'ABC123'
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['dna_sequence']['fmin'] = '42'
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['dna_sequence']['fmax'] = '4242'
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['protein_sequence'] = {}
        d['model_sequences']['sequence']['foo']['protein_sequence']['GI'] = '123456789'
        self.assertEqual([], card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        d['model_sequences']['sequence']['foo']['protein_sequence']['sequence'] = 'III'
        expected = [('foo', '123456789', 'ABC123', '42', '4242', 'ACGT', 'III')]
        self.assertEqual(expected, card_record.CardRecord._dna_seqs_and_genbank_ids(d))
        del d['model_sequences']['sequence']['foo']['protein_sequence']['GI']
        d['model_sequences']['sequence']['foo']['protein_sequence']['accession'] = '123456789'
        self.assertEqual(expected, card_record.CardRecord._dna_seqs_and_genbank_ids(d))

        d['model_sequences']['sequence']['bar'] = {
            'dna_sequence': {
                'sequence': 'TTTT',
                'fmin': '42',
                'fmax': '4242',
                'accession': 'BCD234',
            },
            'protein_sequence': {
                'GI': '',
                'sequence': ''
            }
        }

        expected = [('bar', 'NA', 'BCD234', '42', '4242', 'TTTT', '')] + expected
        self.assertEqual(expected, card_record.CardRecord._dna_seqs_and_genbank_ids(d))


    def test_snps(self):
        '''test snps'''
        d = {'spam': 'eggs'}
        self.assertEqual(set(), card_record.CardRecord._snps(d))
        d['model_param'] = {}
        self.assertEqual(set(), card_record.CardRecord._snps(d))
        d['model_param']['snp'] = {}
        self.assertEqual(set(), card_record.CardRecord._snps(d))
        d['model_param']['snp']['param_value'] = {}
        self.assertEqual(set(), card_record.CardRecord._snps(d))
        d['model_param']['snp']['param_value'] = {
                '1': 'I42L',
                '2': 'S100T',
       }
        self.assertEqual({'I42L', 'S100T'}, card_record.CardRecord._snps(d))


    def test_get_data(self):
        d = {
            'ARO_id': '123',
            'ARO_accession': '1234567',
            'ARO_name': 'ARO_name1',
            'ARO_description': 'ARO description that we want.',
            'model_id': '1',
            'model_name': 'Model_name1',
            'model_type': 'protein homolog model',
            'model_type_id': '12345',
            'model_description': 'Models to detect proteins conferring antibiotic resistance, which include a reference protein sequence and a curated BLASTP cut-off.',
            'model_sequences': {
                'sequence': {
                    '1234': {
                        'protein_sequence': {
                            'sequence': 'MCDE*',
                            'GI': '229597524'
                        },
                        'dna_sequence': {
                            'sequence': 'ATGTGCGATGAATAA',
                            'strand': '+',
                            'fmax': '1194',
                            'fmin': '0',
                            'accession': 'XX0000001'
                        },
                        'NCBI_taxonomy': {
                            'NCBI_taxonomy_cvterm_id': '234567',
                            'NCBI_taxonomy_id': '42',
                            'NCBI_taxonomy_name': 'Genus1 species1'
                        }
                    }
                }
            },
            'model_param': {
                'blastp_evalue': {} # we're ignoring this, so make it empty for tests to save a few lines
            },
            'ARO_category': {
                '36696': {
                    'category_aro_description': 'Enzyme that catalyzes the inactivation of an antibiotic resulting in resistance.  Inactivation includes chemical modification, destruction, etc.',
                    'category_aro_cvterm_id': '36696',
                    'category_aro_accession': '3000557',
                    'category_aro_name': 'antibiotic inactivation enzyme'
                },
                '36268': {
                    'category_aro_description': 'Genes conferring resistance to beta-lactams.',
                    'category_aro_cvterm_id': '36268',
                    'category_aro_accession': '3000129',
                    'category_aro_name': 'beta-lactam resistance gene'
                }
            },
        }

        expected = {
            'ARO_id': '123',
            'ARO_accession': '1234567',
            'ARO_name': 'ARO_name1',
            'ARO_description': 'ARO description that we want.',
            'dna_seqs_and_ids': [(
                '1234',
                '229597524',
                'XX0000001',
                '0',
                '1194',
                'ATGTGCGATGAATAA',
                'MCDE*'
            )],
            'snps': set(),
        }

        record = card_record.CardRecord(d)
        got = record.get_data()
        self.assertEqual(expected, got)

        d['model_param'] = {
            'snp': {
                'param_value': {
                    '1': 'I42L',
                    '2': 'S100T',
                }
            }
        }

        expected['snps'] = {'I42L', 'S100T'}
        record = card_record.CardRecord(d)
        got = record.get_data()
        self.assertEqual(expected, got)




import unittest
import os
import filecmp
from ariba import mlst_reporter

modules_dir = os.path.dirname(os.path.abspath(mlst_reporter.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestMlstReporter(unittest.TestCase):
    def test_all_present_perfect(self):
        '''test when all alleles 100% there no variants'''
        profile_in = os.path.join(data_dir, 'mlst_reporter.profile.in.tsv')
        report_in = os.path.join(data_dir, 'mlst_reporter.all_present_perfect.report.in.tsv')
        expected_out_simple = os.path.join(data_dir, 'mlst_reporter.all_present_perfect.report.out.tsv')
        expected_out_all = os.path.join(data_dir, 'mlst_reporter.all_present_perfect.report.out.details.tsv')
        tmp_out = 'tmp.mlst_reporter.test_all_present_perfect'
        got_simple = tmp_out + '.tsv'
        got_details = tmp_out + '.details.tsv'
        reporter = mlst_reporter.MlstReporter(report_in, profile_in, tmp_out)
        reporter.run()
        self.assertTrue(filecmp.cmp(expected_out_simple, got_simple, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out_all, got_details, shallow=False))
        os.unlink(got_simple)
        os.unlink(got_details)


    def test_new_set(self):
        '''test when allele combination is new'''
        profile_in = os.path.join(data_dir, 'mlst_reporter.profile.in.tsv')
        report_in = os.path.join(data_dir, 'mlst_reporter.new_st.report.in.tsv')
        expected_out_simple = os.path.join(data_dir, 'mlst_reporter.new_st.report.out.tsv')
        expected_out_all = os.path.join(data_dir, 'mlst_reporter.new_st.report.out.details.tsv')
        tmp_out = 'tmp.mlst_reporter.test_new_st'
        got_simple = tmp_out + '.tsv'
        got_details = tmp_out + '.details.tsv'
        reporter = mlst_reporter.MlstReporter(report_in, profile_in, tmp_out)
        reporter.run()
        self.assertTrue(filecmp.cmp(expected_out_simple, got_simple, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out_all, got_details, shallow=False))
        os.unlink(got_simple)
        os.unlink(got_details)


    def test_one_gene_missing(self):
        '''test when one gene missing'''
        profile_in = os.path.join(data_dir, 'mlst_reporter.profile.in.tsv')
        report_in = os.path.join(data_dir, 'mlst_reporter.one_gene_missing.in.tsv')
        expected_out_simple = os.path.join(data_dir, 'mlst_reporter.one_gene_missing.out.tsv')
        expected_out_all = os.path.join(data_dir, 'mlst_reporter.one_gene_missing.out.details.tsv')
        tmp_out = 'tmp.mlst_reporter.test_one_gene_missing'
        got_simple = tmp_out + '.tsv'
        got_details = tmp_out + '.details.tsv'
        reporter = mlst_reporter.MlstReporter(report_in, profile_in, tmp_out)
        reporter.run()
        self.assertTrue(filecmp.cmp(expected_out_simple, got_simple, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out_all, got_details, shallow=False))
        os.unlink(got_simple)
        os.unlink(got_details)


    def test_het_snp(self):
        '''test when one gene has a het snp'''
        profile_in = os.path.join(data_dir, 'mlst_reporter.profile.in.tsv')
        report_in = os.path.join(data_dir, 'mlst_reporter.het_snp.in.tsv')
        expected_out_simple = os.path.join(data_dir, 'mlst_reporter.het_snp.out.tsv')
        expected_out_all = os.path.join(data_dir, 'mlst_reporter.het_snp.out.details.tsv')
        tmp_out = 'tmp.mlst_reporter.test_het_snp'
        got_simple = tmp_out + '.tsv'
        got_details = tmp_out + '.details.tsv'
        reporter = mlst_reporter.MlstReporter(report_in, profile_in, tmp_out)
        reporter.run()
        self.assertTrue(filecmp.cmp(expected_out_simple, got_simple, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out_all, got_details, shallow=False))
        os.unlink(got_simple)
        os.unlink(got_details)


    def test_het_snps(self):
        '''test when one gene has two het snps'''
        profile_in = os.path.join(data_dir, 'mlst_reporter.profile.in.tsv')
        report_in = os.path.join(data_dir, 'mlst_reporter.het_snps.in.tsv')
        expected_out_simple = os.path.join(data_dir, 'mlst_reporter.het_snps.out.tsv')
        expected_out_all = os.path.join(data_dir, 'mlst_reporter.het_snps.out.details.tsv')
        tmp_out = 'tmp.mlst_reporter.test_het_snps'
        got_simple = tmp_out + '.tsv'
        got_details = tmp_out + '.details.tsv'
        reporter = mlst_reporter.MlstReporter(report_in, profile_in, tmp_out)
        reporter.run()
        self.assertTrue(filecmp.cmp(expected_out_simple, got_simple, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out_all, got_details, shallow=False))
        os.unlink(got_simple)
        os.unlink(got_details)



import csv
import json
import os
import re
import sys
import tempfile

from Bio import SeqIO
import pyfastaq

from ariba import common, flag, ref_preparer

data_dir = os.path.join(os.path.dirname(__file__), 'tb_data')
assert os.path.exists(data_dir)


def report_to_resistance_dict(infile):
    '''Takes final ariba report.tsv file, and extracts
    resistance calls, returning a dict of
    drug name -> list of mutations.
    each "mutation" in the list is a tuple of (gene name, mutation).
    Mutation is of the form X42Y, or "incomplete_gene" for katG and
    pncA when they are not complete.
    This all assumes that the reference data are in the "correct"
    form, where the variant descriptions in the var_description column of the
    TSV file ends with a comma-separated list of the drug names'''
    complete_genes = {'katG': 'Isoniazid', 'pncA': 'Pyrazinamide'}
    res_calls = {}
    incomplete_genes = set()
    with open(infile) as f:
        reader = csv.DictReader(f, delimiter='\t')
        for d in reader:
            if d['ref_name'] in complete_genes and d['gene'] == '1':
                f = flag.Flag(int(d['flag']))
                if not f.has('complete_gene'):
                    incomplete_genes.add(d['ref_name'])

            if d['has_known_var'] == '1':
                if 'Original mutation' in d['var_description']:
                    drugs = d['var_description'].split(':')[-1].split('.')[0].split()[-1].split(',')
                    change = d['var_description'].split()[-1]
                else:
                    drugs = d['var_description'].split()[-1].split(',')
                    change = d['known_var_change']
                for drug in drugs:
                    if drug not in res_calls:
                        res_calls[drug] = []
                    res_calls[drug].append((d['ref_name'], change))

    for gene in incomplete_genes:
        drug = complete_genes[gene]
        if drug not in res_calls:
            res_calls[drug] = []
        res_calls[drug].append((gene, 'Incomplete_gene'))

    return res_calls


def genbank_to_gene_coords(infile, genes):
    '''Input file in genbank format. genes = list of gene names to find.
    Returns dict of gene name -> {start: x, end: y}, where x and y are
    zero-based. x<y iff gene is on forwards strand'''
    coords = {}

    for seq_record in SeqIO.parse(infile, "genbank"):
        for feature in seq_record.features:
            if feature.type == 'gene':
                gene_name = feature.qualifiers.get('gene', [None])[0]
                if gene_name not in genes:
                    continue

                if feature.location.strand == 1:
                    coords[gene_name] = {'start': int(feature.location.start), 'end': int(feature.location.end) - 1}
                else:
                    coords[gene_name] = {'end': int(feature.location.start), 'start': int(feature.location.end) - 1}

    return coords


def load_mutations(gene_coords, mutation_to_drug_json, variants_txt, upstream_before=100):
    '''Load mutations from "mykrobe-style" files. mutation_to_drug_json is json file
    of mutation -> list of drugs. variants_txt is text file of variants used my mykrobe's
    make probes. gene_coords should be dict of gene coords made by the function
    genbank_to_gene_coords'''
    with open(mutation_to_drug_json) as f:
        drug_data = json.load(f)

    mutations = []
    genes_with_indels = set()
    genes_need_upstream = set()
    genes_non_upstream = set()

    with open(variants_txt) as f:
        for line in f:
            gene, variant, d_or_p = line.rstrip().split('\t')
            coding = 0 if gene == 'rrs' else 1
            d = {'gene': gene, 'var': variant, 'coding': coding, 'upstream': False}
            drug_data_key = d['gene'] + '_' + d['var']
            if drug_data_key not in drug_data:
                print('KEY', drug_data_key, 'NOT FOUND', file=sys.stderr)
            else:
                d['drugs'] = ','.join(sorted(drug_data[drug_data_key]))

            if d_or_p == 'DNA' and gene != 'rrs':
                assert gene != 'rrs'
                re_match = re.match('([ACGT]+)(-?[0-9]+)([ACGTX]+)', d['var'])
                try:
                    ref, pos, alt = re_match.groups()
                except:
                    print('regex error:', d['var'], file=sys.stderr)
                    continue

                pos = int(pos)
                if len(ref) != len(alt):
                    genes_with_indels.add(d['gene'])
                    continue
                elif pos > 0:
                    #print('ignoring synonymous change (not implemented):', d['gene'], d['var'], d['drugs'], file=sys.stderr)
                    continue
                elif pos < 0:
                    this_gene_coords = gene_coords[d['gene']]
                    d['upstream'] = True
                    if this_gene_coords['start'] < this_gene_coords['end']:
                        variant_pos_in_output_seq = upstream_before + pos + 1
                    else:
                        variant_pos_in_output_seq = upstream_before + pos + 1
                    assert variant_pos_in_output_seq > 0
                    d['var'] = ref + str(variant_pos_in_output_seq) + alt
                    d['original_mutation'] = variant
                    genes_need_upstream.add(d['gene'])
                elif pos == 0:
                    print('Zero coord!', d, file=sys.stderr)
                    continue
                else:
                    print('deal with?', d, file=sys.stderr)
                    continue

            mutations.append(d)
            if not d['upstream']:
                genes_non_upstream.add(d['gene'])

    return mutations, genes_with_indels, genes_need_upstream, genes_non_upstream


def write_prepareref_fasta_file(outfile, gene_coords, genes_need_upstream, genes_non_upstream, upstream_before=100, upstream_after=100):
    '''Writes fasta file to be used with -f option of prepareref'''
    tmp_dict = {}
    fasta_in = os.path.join(data_dir, 'NC_000962.3.fa.gz')
    pyfastaq.tasks.file_to_dict(fasta_in, tmp_dict)
    ref_seq = tmp_dict['NC_000962.3']

    with open(outfile, 'w') as f:
        for gene in genes_non_upstream:
            start = gene_coords[gene]['start']
            end = gene_coords[gene]['end']
            if start < end:
                gene_fa = pyfastaq.sequences.Fasta(gene, ref_seq[start:end+1])
            else:
                gene_fa = pyfastaq.sequences.Fasta(gene, ref_seq[end:start+1])
                gene_fa.revcomp()

            print(gene_fa, file=f)

        for gene in genes_need_upstream:
            start = gene_coords[gene]['start']
            end = gene_coords[gene]['end']
            if start < end:
                gene_fa = pyfastaq.sequences.Fasta(gene, ref_seq[start - upstream_before:start + upstream_after])
            else:
                gene_fa = pyfastaq.sequences.Fasta(gene, ref_seq[start - upstream_after + 1:start + upstream_before + 1])
                gene_fa.revcomp()

            gene_fa.id += '_upstream'
            print(gene_fa, file=f)


def write_prepareref_metadata_file(mutations, outfile):
    aa_letters = {'G', 'P', 'A', 'V', 'L', 'I', 'M', 'C', 'F', 'Y',
        'W', 'H', 'K', 'R', 'Q', 'N', 'E', 'D', 'S', 'T'}

    with open(outfile, 'w') as f:
        for d in mutations:
            if d['upstream']:
                gene = d['gene'] + '_upstream'
                coding = '0'
            else:
                gene = d['gene']
                coding = d['coding']

            if 'original_mutation' in d:
                original_mutation_string = '. Original mutation ' + d['original_mutation']
            else:
                original_mutation_string = ''

            if d['var'].endswith('X'):
                ref = d['var'][0]
                if d['coding'] == 1 and not d['upstream']:
                    letters = aa_letters
                else:
                    letters = {'A', 'C', 'G', 'T'}

                assert ref in letters

                for x in letters:
                    if x == ref:
                        continue
                    variant = d['var'][:-1] + x
                    print(gene, coding, 1, variant, '.', 'Resistance to ' + d['drugs'] + original_mutation_string, sep='\t', file=f)
            else:
                print(gene, coding, 1, d['var'], '.', 'Resistance to ' + d['drugs'] + original_mutation_string, sep='\t', file=f)


def make_prepareref_files(outprefix):
    genbank_file = os.path.join(data_dir, 'NC_000962.3.gb')
    mut_to_drug_json = os.path.join(data_dir, 'panel.20190102.json')
    panel_txt_file = os.path.join(data_dir, 'panel.20190102.txt')
    fasta_out = outprefix + '.fa'
    tsv_out = outprefix + '.tsv'

    with open(panel_txt_file) as f:
        genes = set([x.split()[0] for x in f])

    ref_gene_coords = genbank_to_gene_coords(genbank_file, genes)
    mutations, genes_with_indels, genes_need_upstream, genes_non_upstream = load_mutations(ref_gene_coords, mut_to_drug_json, panel_txt_file)
    write_prepareref_fasta_file(fasta_out, ref_gene_coords, genes_need_upstream, genes_non_upstream)
    write_prepareref_metadata_file(mutations, tsv_out)


def make_prepareref_dir(outdir):
    if os.path.exists(outdir):
        raise Exception('Output directory ' + outdir + ' already exists. Cannot continue')

    tmpdir = tempfile.mkdtemp(prefix=outdir + '.tmp', dir=os.getcwd())
    tmp_prefix = os.path.join(tmpdir, 'out')
    make_prepareref_files(tmp_prefix)
    ref_prep = ref_preparer.RefPreparer(
        [tmp_prefix + '.fa'],
        None,
        metadata_tsv_files=[tmp_prefix + '.tsv'],
        run_cdhit=False,
        threads=1,
    )
    ref_prep.run(outdir)
    common.rmtree(tmpdir)

    json_data = {'tb': True}
    json_file = os.path.join(outdir, '00.params.json')
    with open(json_file, 'w') as f:
        json.dump(json_data, f)



import unittest
import os
import filecmp
from ariba import ref_genes_getter

modules_dir = os.path.dirname(os.path.abspath(ref_genes_getter.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestRefGenesGetter(unittest.TestCase):
    def test_fix_virulencefinder_fasta_file(self):
        '''test _fix_virulencefinder_fasta_file'''
        infile = os.path.join(data_dir, 'ref_genes_getter.fix_virulencefinder_fasta_file.in.fa')
        tmp_file = 'tmp.test.ref_genes_getter.fix_virulencefinder_fasta_file.out.fa'
        expected_file = os.path.join(data_dir, 'ref_genes_getter.fix_virulencefinder_fasta_file.out.fa')
        ref_genes_getter.RefGenesGetter._fix_virulencefinder_fasta_file(infile, tmp_file)
        self.assertTrue(filecmp.cmp(expected_file, tmp_file, shallow=False))
        os.unlink(tmp_file)



import csv
import sys
import random
import re
import os
import itertools
import collections
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.cm as cmx
import math
import pyfastaq
from ariba import reference_data

class Error (Exception): pass

regex_string_to_float = re.compile(r'\s*(?P<lt_or_gt>[<>]?)\s*(?P<equals>=?)\s*(?P<number>[0-9.]+)\s*$')

regex_position_from_var = re.compile(r'^[^0-9]*(?P<coord>[0-9]+)[^0-9]*$')

class MicPlotter:
    def __init__(self,
      refdata_dir,
      antibiotic,
      mic_file,
      summary_file,
      outprefix,
      use_hets='yes',
      main_title=None,
      plot_height=7,
      plot_width=7,
      log_y=2,
      plot_types="points,violin",
      jitter_width=0.1,
      no_combinations=False,
      hlines='',
      point_size=4,
      point_scale=1,
      dot_size=100,
      dot_outline=False,
      dot_y_text_size=7,
      panel_heights='9,2',
      panel_widths='5,1',
      colourmap='Accent',
      number_of_colours=0,
      colour_skip=None,
      interrupted=False,
      violin_width=0.75,
      xkcd=False,
      min_samples=1,
      count_legend_x=-2,
      out_format='pdf',
      p_cutoff=0.05
    ):
        refdata_fa = os.path.join(refdata_dir, '02.cdhit.all.fa')
        refdata_tsv = os.path.join(refdata_dir, '01.filter.check_metadata.tsv')
        self.refdata = reference_data.ReferenceData([refdata_fa], [refdata_tsv])
        self.antibiotic = antibiotic
        self.mic_file = mic_file
        self.summary_file = summary_file
        self.outprefix = outprefix

        allowed_use_hets = {'yes', 'no', 'exclude'}
        if not use_hets in allowed_use_hets:
            raise Error('Error in use_hets option. Allowed options are: ' + str(allowed_use_hets) + '. Got: ' + use_hets)
        self.use_hets = use_hets

        self.main_title = self.antibiotic if main_title is None else main_title
        self.plot_height = plot_height
        self.plot_width = plot_width
        self.log_y = log_y
        self.plot_types = set(plot_types.split(','))

        allowed_plot_types = {'point', 'violin'}
        if not self.plot_types.issubset(allowed_plot_types):
            raise Error('Error in plot_types option. Allowed types are: ' + str(allowed_plot_types) + '. Got: ' +  str(self.plot_types))

        self.jitter_width = jitter_width
        self.no_combinations = no_combinations

        try:
            if len(hlines) == 0:
                self.hlines = []
            else:
                self.hlines = [float(x) for x in hlines.split(',')]
        except:
            raise Error('Error in hlines option. Needs to be a list of numbers separated by commas, or empty. Got this:\n' + hlines)

        self.point_size = point_size
        self.point_scale = point_scale
        self.dot_size = dot_size
        self.dot_outline = dot_outline
        self.dot_y_text_size = dot_y_text_size

        try:
            self.panel_heights = [int(x) for x in panel_heights.split(',')]
        except:
            raise Error('Error in panel_heights option. Needs to be of the form integer1,integer2. Got this:\n' + panel_heights)

        try:
            self.panel_widths = [int(x) for x in panel_widths.split(',')]
        except:
            raise Error('Error in panel_widths option. Needs to be of the form integer1,integer2. Got this:\n' + panel_widths)

        self.colourmap = colourmap
        self.number_of_colours = number_of_colours

        if colour_skip is None:
            self.colour_skip = None
        else:
            try:
                self.colour_skip = [float(x) for x in colour_skip.split(',')]
            except:
                raise Error('Error in colour_skip option. Needs to be of the form a,b where 0 <= a < b <= 1. Got this:\n' + colour_skip)

        self.interrupted = interrupted
        self.violin_width = violin_width
        if xkcd:
            plt.xkcd()
        self.min_samples = min_samples
        self.count_legend_x = count_legend_x
        self.out_format = out_format
        self.p_cutoff = p_cutoff


    @classmethod
    def _mic_string_to_float(cls, s):
        regex_match = regex_string_to_float.match(s)

        if regex_match is None or regex_match.group('number') == '.':
            if s.strip() in {'NA', 'na', '', '.'}:
                return 'NA'
            else:
                return None

        try:
            flt = float(regex_match.group('number'))
        except:
            return None

        if regex_match.group('equals') == '':
            if regex_match.group('lt_or_gt') == '<':
                return 0.5 * flt
            elif regex_match.group('lt_or_gt') == '>':
                return 2 * flt

        return flt


    @classmethod
    def _load_mic_file(cls, infile):
        mic_data = {}

        with open(infile) as f:
            reader = csv.DictReader(f, delimiter='\t')
            if reader.fieldnames[0] != 'Sample':
                raise Error('Error. Expected first column of MIC file "' + infile + '" to be "Sample"')

            for row in reader:
                mic_data[row['Sample']] = {x: MicPlotter._mic_string_to_float(row[x]) for x in reader.fieldnames[1:]}

        return mic_data


    @classmethod
    def _load_summary_file(cls, infile):
        data = {}

        with open(infile) as f:
            reader = csv.DictReader(f, delimiter=',')

            if reader.fieldnames[0] != 'name':
                raise Error('Error. Expected first column of summary file "' + infile + '" to be "name"')

            clusters = [x.split('.', maxsplit=1)[0] for x in reader.fieldnames[1:]]

            for row in reader:
                data[row['name']] = {}

                for field in row:
                    if field == 'name':
                        continue

                    cluster, col = field.split('.', maxsplit=1)
                    if cluster not in clusters:
                        raise Error('Cluster "' + cluster + '" not recognised. Cannot continue')
                    if cluster not in data[row['name']]:
                        data[row['name']][cluster] = {}

                    try:
                        value = float(row[field])
                    except:
                        value = row[field]
                    data[row['name']][cluster][col] = value

        return data


    @classmethod
    def _get_colours(cls, total_length, number_of_colours, colormap, skip=None):
        if number_of_colours == 1:
            return ["black"] * total_length
        elif number_of_colours == 0:
            cmap = cmx.get_cmap(colormap)
            if skip is None:
                vals = [1.0 * x / (total_length - 1) for x in range(total_length)]
            else:
                assert len(skip) == 2 and 0 <= skip[0] <= 1 and 0 <= skip[1] <= 1
                if skip[-1] == 1:
                    vals = [skip[0] * x / (total_length - 1) for x in range(total_length)]
                elif skip[0] == 0:
                    vals = [skip[1] + (1 - skip[1]) * x / (total_length - 1) for x in range(total_length)]
                else:
                    length = 1 - (skip[1] - skip[0])
                    vals = [(length) * x / (total_length - 1) for x in range(total_length)]
                    vals = [x if x < skip[0] else x + (1-length) for x in vals]

            return [cmap(x) for x in vals]
        else:
            cmap = cmx.get_cmap(colormap)
            colours = []
            for i in itertools.cycle(range(number_of_colours)):
                colours.append(cmap(i))
                if len(colours) >= total_length:
                    break
            return colours


    @classmethod
    def _get_top_plot_data(cls, summary_data, mic_data, antibiotic, use_hets, refdata=None, no_combinations=False, interrupted=False, outfile=None):
        assert use_hets in {'yes', 'no', 'exclude'}
        if outfile is not None:
            f = pyfastaq.utils.open_file_write(outfile)
            print('Sample\tMIC\tMutations', file=f)

        ignore_columns = {'assembled', 'match', 'ref_seq', 'pct_id', 'known_var', 'novel_var', 'MULTIPLE'}
        all_mutations = set()
        all_mutations_seen_combinations = set()
        top_plot_data = {} # cluster combination -> list of y coords (MIC values)

        for sample in sorted(summary_data):
            if sample not in mic_data:
                raise Error('No MIC data found for sample "' + sample + '". Cannot continue')
            if antibiotic not in mic_data[sample]:
                raise Error('Antibiotic "' + antibiotic + '" not found. Cannot continue')

            if mic_data[sample][antibiotic] == 'NA':
                continue

            mutations = set()
            found_het_and_exclude = False

            for cluster in summary_data[sample]:
                if 'assembled' in summary_data[sample][cluster] and summary_data[sample][cluster]['assembled'] == 'interrupted' and interrupted:
                    mutations.add(cluster + '.interrupted')

                if refdata is not None and 'match' in summary_data[sample][cluster] and summary_data[sample][cluster]['match'] == 'yes' and 'ref_seq' in summary_data[sample][cluster]:
                    ref_type, variant_only = refdata.sequence_type(summary_data[sample][cluster]['ref_seq'])
                    if not variant_only:
                        mutations.add(cluster + '.present')

                for column, value in summary_data[sample][cluster].items():
                    if column in ignore_columns or column.endswith('.%'):
                        continue

                    if value == 'yes' or (use_hets == 'yes' and value == 'het'):
                        mutations.add(cluster + '.' + column.strip())
                    elif use_hets == 'exclude' and value == 'het':
                        found_het_and_exclude = True
                        break

                if found_het_and_exclude:
                    break

            if found_het_and_exclude:
                continue


            if len(mutations) == 0:
                mutations.add('without_mutation')

            all_mutations.update(mutations)
            mutations = list(mutations)
            mutations.sort()
            if no_combinations:
                for mutation in mutations:
                    all_mutations_seen_combinations.add((mutation,))
                    if mutation not in top_plot_data:
                        top_plot_data[mutation] = []
                    top_plot_data[mutation].append(mic_data[sample][antibiotic])
                    if outfile is not None:
                        print(sample, mic_data[sample][antibiotic], mutation, sep='\t', file=f)
            else:
                all_mutations_seen_combinations.add(tuple(mutations))
                mutations = '.'.join(mutations)
                if mutations not in top_plot_data:
                    top_plot_data[mutations] = []
                top_plot_data[mutations].append(mic_data[sample][antibiotic])
                if outfile is not None:
                    print(sample, mic_data[sample][antibiotic], mutations, sep='\t', file=f)


        if outfile is not None:
            pyfastaq.utils.close(f)

        return top_plot_data, all_mutations, all_mutations_seen_combinations


    @classmethod
    def _filter_top_plot_data(cls, top_plot_data, all_mutations, seen_combinations, min_samples):
        if min_samples == 1:
            return top_plot_data, all_mutations, seen_combinations

        new_top_plot_data = {}
        new_all_mutations = set()
        new_seen_combinations = set()

        for mutation_tuple in seen_combinations:
            mutation_string = '.'.join(mutation_tuple)
            mics = top_plot_data[mutation_string]

            if len(mics) >= min_samples:
                new_top_plot_data[mutation_string] = mics
                new_seen_combinations.add(mutation_tuple)
                new_all_mutations.update(mutation_tuple)

        return new_top_plot_data, new_all_mutations, new_seen_combinations


    @classmethod
    def _top_plot_y_ticks(cls, mic_data, antibiotic, log_y):
        mic_values = set()
        for sample in mic_data:
            mic = mic_data[sample][antibiotic]
            if mic not in [None, 'NA']:
                mic_values.add(mic)

        max_mic = max(mic_values)
        min_mic = min(mic_values)
        new_mic_values = []
        i = 1
        while i < max_mic * 2:
            new_mic_values.append(i)
            i *= 2

        i = 0.5
        while i > min_mic / 2:
            new_mic_values.append(i)
            i *= 0.5

        new_mic_values.sort()
        new_mic_values = [round(x, 4) for x in new_mic_values]

        if log_y > 0:
            tick_positions = [math.log(x, log_y) for x in new_mic_values]
        else:
            tick_positions = new_mic_values

        return tick_positions, new_mic_values


    @classmethod
    def _top_plot_scatter_counts(cls, mutations, top_plot_data, colours, log_y):
        x_coords = []
        y_coords = []
        sizes = []
        colour_list = []

        for i, mutation in enumerate(mutations):
            counts = collections.Counter(top_plot_data[mutation])
            for mic in sorted(counts):
                x_coords.append(i + 1)
                if log_y > 0:
                    y_coords.append(math.log(mic, log_y))
                else:
                    y_coords.append(mic)
                sizes.append(counts[mic])
                colour_list.append(colours[i])

        return x_coords, y_coords, sizes, colour_list


    @classmethod
    def _top_plot_scatter_data(cls, mutations, top_plot_data, colours, log_y, x_jitter):
        x_coords = []
        y_coords = []
        colour_list = []

        for i, mutation in enumerate(mutations):
            for mic in top_plot_data[mutation]:
                if len(top_plot_data[mutation]) > 1:
                    x_coords.append(i + 1 + random.uniform(-x_jitter, x_jitter))
                else:
                    x_coords.append(i + 1)

                if log_y > 0:
                    y_coords.append(math.log(mic, log_y))
                else:
                    y_coords.append(mic)
                colour_list.append(colours[i])

        return x_coords, y_coords, colour_list


    @classmethod
    def _top_plot_violin_data(cls, mutations, top_plot_data, log_y):
        violin_data = []
        violin_pos = []

        for i, mutation in enumerate(mutations):
            if log_y > 0:
                violin_data.append([math.log(x, log_y) for x in top_plot_data[mutation]])
            else:
                violin_data.append(top_plot_data[mutation])
            violin_pos.append(i + 1)

        return violin_data, violin_pos


    @classmethod
    def _ordered_bottom_plot_rows(cls, mutations):
        l = []
        infinity = float('inf')

        for x in mutations:
            try:
                cluster, variant = x.split('.', maxsplit=1)
            except:
                l.append((x, infinity, x))
                continue

            if '.' in variant:
                try:
                    var_group, var = variant.split('.', maxsplit=1)
                except:
                    var_group = None
                    var = variant

                variant = var

            regex_match = regex_position_from_var.match(variant)
            if regex_match is not None and regex_match.group('coord') != '':
                coord = int(regex_match.group('coord'))
            else:
                coord = infinity

            l.append((cluster, coord, x))

        l.sort()
        return [x[-1] for x in l]


    @classmethod
    def _ordered_columns(cls, mutations, top_plot_data):
        # FIXME
        return sorted(list(mutations))


    @classmethod
    def _bottom_scatter_data(cls, bottom_plot_rows, columns, colours, outline=False):
        x_coords = []
        y_coords = []
        colour_list = []

        for i, row in enumerate(bottom_plot_rows):
            for j, col in enumerate(columns):
                if row in col:
                    x_coords.append(j + 1)
                    y_coords.append(len(bottom_plot_rows) - i)
                    colour_list.append(colours[j])
                elif outline:
                    x_coords.append(j + 1)
                    y_coords.append(len(bottom_plot_rows) - i)
                    colour_list.append("white")

        return x_coords, y_coords, colour_list


    @classmethod
    def _right_plot_data(cls, scatter_count_sizes, x_pos):
        y_max = max(scatter_count_sizes)
        if y_max > 100:
            y_max = int(math.ceil(y_max / 100.0)) * 100
            sizes = [5, 50]
        else:
            y_max = int(math.ceil(y_max / 10.0)) * 10
            sizes = [5, 10]

        while sizes[-1] < y_max:
            sizes.append(sizes[-1]*2)

        x_coords = [x_pos] * len(sizes)
        y_coords = [x + 1 for x in range(len(sizes))]
        y_coords.reverse()
        return x_coords, y_coords, sizes


    @classmethod
    def _pairwise_compare(cls, violin_data, columns, outfile, p_cutoff, compare_test):
        try:
            import scipy.stats
        except:
            print('WARNING: skipping Mann Whitney tests because scipy.stats not found', file=sys.stderr)
            return

        output = []

        for i, list1 in enumerate(violin_data):
            for j, list2 in enumerate(violin_data):
                if j <= i or len(list1) < 2 or len(list2) < 2:
                    continue

                list1set = set(list1)

                if len(list1set) == 1 and list1set == set(list2):
                    statistic = 'NA'
                    pvalue = 1
                else:
                    if compare_test == 'mannwhitneyu':
                        statistic, pvalue = scipy.stats.mannwhitneyu(list1, list2, alternative='two-sided')
                    elif compare_test == 'ks_2samp':
                        statistic, pvalue = scipy.stats.ks_2samp(list1, list2)
                    else:
                        raise Error('Test "' + compare_test + '" not recognised. Cannot continue')

                effect_size = abs(scipy.stats.norm.ppf(pvalue) / math.sqrt(len(list1) + len(list2)))
                significant = 'yes' if pvalue < p_cutoff else 'no'
                output.append((columns[i], columns[j], len(list1), len(list2), pvalue, significant, effect_size))

        output.sort(key=lambda x: x[4])

        with open(outfile, 'w') as f:
            print('Combination1', 'Combination2', 'Size1', 'Size2', 'p-value', 'significant', 'effect_size', 'corrected_p-value', 'corrected_significant', 'corrected_effect_size', sep='\t', file=f)
            for x in output:
                corrected_p = min(1, len(output) * x[4])
                corrected_significant = 'yes' if corrected_p < p_cutoff else 'no'
                corrected_effect_size = scipy.stats.norm.ppf(corrected_p) / math.sqrt(x[2] + x[3])
                print('\t'.join([str(z) for z in x]), corrected_p, corrected_significant, corrected_effect_size, sep='\t', file=f)


    def _make_plot(self, mic_data, top_plot_data, all_mutations, mut_combinations):
        bottom_plot_rows = MicPlotter._ordered_bottom_plot_rows(all_mutations)
        columns = MicPlotter._ordered_columns(mut_combinations, top_plot_data)
        colours = MicPlotter._get_colours(len(columns), self.number_of_colours, self.colourmap, self.colour_skip)
        bottom_scatter_x, bottom_scatter_y, bottom_colours = MicPlotter._bottom_scatter_data(bottom_plot_rows, columns, colours, outline=self.dot_outline)
        columns = ['.'.join(x) for x in columns]
        assert len(colours) == len(columns)
        max_x = len(colours) + 1

        scatter_count_x, scatter_count_y, scatter_count_sizes, scatter_count_colours = MicPlotter._top_plot_scatter_counts(columns, top_plot_data, colours, self.log_y)
        scatter_data_x, scatter_data_y, scatter_data_colours = MicPlotter._top_plot_scatter_data(columns, top_plot_data, colours, self.log_y, self.jitter_width)
        violin_data, violin_positions = MicPlotter._top_plot_violin_data(columns, top_plot_data, self.log_y)
        MicPlotter._pairwise_compare(violin_data, columns, self.outprefix + '.mannwhitney.tsv', self.p_cutoff, 'mannwhitneyu')
        MicPlotter._pairwise_compare(violin_data, columns, self.outprefix + '.ks_2sample.tsv', self.p_cutoff, 'ks_2samp')

        # -------------------- SET UP GRID & PLOTS -----------------
        fig=plt.figure(figsize=(self.plot_width, self.plot_height))
        if 'point' not in self.plot_types:
            self.point_size = 42

        if self.point_size == 0:
            gs = gridspec.GridSpec(2, 2, height_ratios=self.panel_heights, width_ratios=self.panel_widths)
        else:
            gs = gridspec.GridSpec(2, 1, height_ratios=self.panel_heights)

        plots=[]
        plots.append(plt.subplot(gs[0]))
        plots.append(plt.subplot(gs[1]))
        if self.point_size == 0:
            plots.append(plt.subplot(gs[2]))
            bottom_plot_index = 2
        else:
            bottom_plot_index = 1

        # ------------------------- TOP PLOT -----------------------
        for h in self.hlines:
            if self.log_y > 0:
                h = math.log(h, self.log_y)
            plots[0].hlines(h, 0, max_x, linestyle='--', linewidth=1, color='black')


        if 'violin' in self.plot_types:
            violins = plots[0].violinplot(violin_data, violin_positions, widths=self.violin_width, showmeans=False, showextrema=False, showmedians=False)
            for x, pc in enumerate(violins['bodies']):
                pc.set_facecolor(colours[x])
                pc.set_edgecolor(colours[x])

        scaled_count_sizes = [self.point_scale * x for x in scatter_count_sizes]

        if 'point' in self.plot_types:
            if self.point_size == 0:
                plots[0].scatter(scatter_count_x, scatter_count_y, s=scaled_count_sizes, c=scatter_count_colours, linewidth=0)
            else:
                plots[0].scatter(scatter_data_x, scatter_data_y, c=scatter_data_colours, s=self.point_size)

        if self.log_y > 0:
            miny = min(scatter_count_y) - 0.5
            maxy = max(scatter_count_y) + 0.5
        else:
            miny = 0
            maxy = 1.05 * max(scatter_count_y)

        plots[0].axis([0,max(bottom_scatter_x) + 1, miny, maxy])

        y_tick_positions, y_tick_labels = MicPlotter._top_plot_y_ticks(mic_data, self.antibiotic, self.log_y)
        plots[0].yaxis.set_ticks(y_tick_positions)
        plots[0].set_yticklabels(y_tick_labels)
        ylabel = r'MIC ($\mu$g/mL)'
        plots[0].set_ylabel(ylabel)
        plots[0].set_xticklabels([])
        plots[0].set_title(self.main_title, fontsize=18)

        # ------------------------- BOTTOM PLOT -----------------------
        edgecolor = "black" if self.dot_outline else bottom_colours
        plots[bottom_plot_index].axis([0,max(bottom_scatter_x) + 1,0,max(bottom_scatter_y) + 1])
        plots[bottom_plot_index].scatter(bottom_scatter_x, bottom_scatter_y, marker='o', s=self.dot_size, c=bottom_colours, edgecolor=edgecolor, lw=1)
        plots[bottom_plot_index].spines["top"].set_visible(False)
        plots[bottom_plot_index].spines["right"].set_visible(False)
        plots[bottom_plot_index].spines["bottom"].set_visible(False)
        plots[bottom_plot_index].spines["left"].set_visible(False)
        plots[bottom_plot_index].yaxis.set_tick_params(length=0)
        plots[bottom_plot_index].xaxis.set_ticks([])
        plots[bottom_plot_index].set_xticklabels([])
        plots[bottom_plot_index].yaxis.set_ticks([(i+1) for i in range(len(bottom_plot_rows))])
        plots[bottom_plot_index].set_yticklabels(bottom_plot_rows[::-1], fontsize=self.dot_y_text_size)

        # ------------------------- RIGHT PLOT -------------------------
        if self.point_size == 0:
            right_x_coord = 0.75
            right_x, right_y, right_sizes = MicPlotter._right_plot_data(scatter_count_sizes, right_x_coord)
            right_scaled_sizes = [self.point_scale * x for x in right_sizes]
            plots[1].scatter(right_x, right_y, s=right_scaled_sizes, c="black")
            plots[1].axis('off')
            plots[1].axis([0,4,-2*len(right_y),len(right_y)+1])
            for i, y in enumerate(right_y):
                plots[1].annotate(right_sizes[i], [right_x_coord + 0.75, y-0.2])
            plots[1].annotate("Counts", [right_x_coord - 0.1, len(right_y) + 0.5])

        plt.tight_layout(w_pad=self.count_legend_x)
        plt.savefig(self.outprefix + '.' + self.out_format)




    def run(self):
        mic_data = MicPlotter._load_mic_file(self.mic_file)
        summary_data = MicPlotter._load_summary_file(self.summary_file)
        boxplot_tsv = self.outprefix + '.data.tsv'
        top_plot_data, all_mutations, combinations = MicPlotter._get_top_plot_data(summary_data, mic_data, self.antibiotic, self.use_hets, refdata=self.refdata, no_combinations=self.no_combinations, interrupted=self.interrupted, outfile=boxplot_tsv)
        top_plot_data, all_mutations, combinations = MicPlotter._filter_top_plot_data(top_plot_data, all_mutations, combinations, self.min_samples)
        self._make_plot(mic_data, top_plot_data, all_mutations, combinations)



import os
import time
import sys
import subprocess
import urllib.request
import pyfastaq


class Error (Exception): pass


def syscall(cmd, allow_fail=False, verbose=False, verbose_filehandle=sys.stdout, print_errors=True, shell=True):
    if verbose:
        print('syscall:', cmd, flush=True, file=verbose_filehandle)
        if not shell:
            print('syscall string:', " ".join('"{}"'.format(_) for _ in cmd), flush=True, file=verbose_filehandle)
    try:
        subprocess.check_output(cmd, shell=shell, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        errors = error.output.decode()
        if print_errors:
            print('The following command failed with exit code', error.returncode, file=sys.stderr)
            print(cmd, file=sys.stderr)
            print('\nThe output was:\n', file=sys.stderr)
            print(errors, file=sys.stderr, flush=True)

        if allow_fail:
            return False, errors
        else:
            sys.exit(1)
    except Exception as msg:
        print("Unexpected exception: ", msg, file=sys.stderr)
        raise
    return True, None


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s


def cat_files(infiles, outfile):
    '''Cats all files in list infiles into outfile'''
    f_out = pyfastaq.utils.open_file_write(outfile)

    for filename in infiles:
        if os.path.exists(filename):
            f_in = pyfastaq.utils.open_file_read(filename)
            for line in f_in:
                print(line, end='', file=f_out)
            pyfastaq.utils.close(f_in)

    pyfastaq.utils.close(f_out)


def download_file(url, outfile, max_attempts=3, sleep_time=2, verbose=False):
    if verbose:
        print('Downloading "', url, '" and saving as "', outfile, '" ...', end='', sep='', flush=True)

    for i in range(max_attempts):
        time.sleep(sleep_time)
        try:
            urllib.request.urlretrieve(url, filename=outfile)
        except:
            continue
        break
    else:
        raise Error('Error downloading: ' + url)

    if verbose:
        print(' done', flush=True)


def rmtree(input_dir):
    '''Does rm -r on input_dir. Meant to replace shutil.rmtree,
    which seems to be causing issues with files not getting deleted
    and the directory non-empty afterwards'''
    syscall('rm -rf ' + input_dir)



import unittest
import filecmp
import os
import pyfastaq
from ariba import common, megares_zip_parser

modules_dir = os.path.dirname(os.path.abspath(megares_zip_parser.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestMegaresZipParser(unittest.TestCase):
    def test_extract_files_ok(self):
        '''test _extract_files when all ok'''
        zip_file = os.path.join(data_dir, 'megares_zip_parse_extract_files_ok.zip')
        tmp_dir = 'tmp.test_megares_extract_files_ok'
        got = megares_zip_parser.MegaresZipParser._extract_files(zip_file, tmp_dir)
        common_dir = os.path.join('megares_zip_parse_extract_files_ok', 'megares_v1.01')
        expected = {
            'annotations': os.path.join(common_dir, 'megares_annotations_v1.01.csv'),
            'fasta': os.path.join(common_dir, 'megares_database_v1.01.fasta'),
            'header_mappings': os.path.join(common_dir, 'megares_to_external_header_mappings_v1.01.tsv')
        }

        self.assertEqual(expected, got)

        for filename in expected.values():
            self.assertTrue(os.path.exists(os.path.join(tmp_dir, filename)))

        common.rmtree(tmp_dir)


    def test_extract_files_one_missing(self):
        '''test _extract_files when one missing'''
        zip_file = os.path.join(data_dir, 'megares_zip_parse_extract_files_one_missing.zip')
        tmp_dir = 'tmp.test_megares_extract_files_one_missing'
        with self.assertRaises(megares_zip_parser.Error):
            got = megares_zip_parser.MegaresZipParser._extract_files(zip_file, tmp_dir)


    def test_load_annotations_file(self):
        '''test _load_annotations_file'''
        infile = os.path.join(data_dir, 'megares_zip_parser_load_annotations.csv')
        expected = {
            'Bla|OXA-1|JN123456|42-141|100|betalactams|Class_A_betalactamases|OXA': {'class': 'betalactams', 'mechanism': 'Class A betalactamases', 'group': 'OXA'},
            'Foo|Bar-1|JN42|1-11|10|foobar|Class_foobar|Bar': {'class': 'Class', 'mechanism': 'foobar', 'group': 'Bar'}
        }
        got = megares_zip_parser.MegaresZipParser._load_annotations_file(infile)
        self.maxDiff = None
        self.assertEqual(expected, got)


    def test_load_header_mappings_file(self):
        '''test _load_header_mappings_file'''
        infile = os.path.join(data_dir, 'megares_zip_parser_load_header_mappings.tsv')
        expected = {
            'Bla|OXA-1|JN123456|42-141|100|betalactams|Class_A_betalactamases|OXA': {'Source_Database': 'SOURCE1', 'Source_Headers(space_separated)': 'source header 1'},
            'Foo|Bar-1|JN42|1-11|10|foobar|Class_foobar|Bar': {'Source_Database': 'SOURCE2', 'Source_Headers(space_separated)': 'source header 2'},
        }
        got = megares_zip_parser.MegaresZipParser._load_header_mappings_file(infile)
        self.maxDiff = None
        self.assertEqual(expected, got)


    def test_write_files(self):
        '''test _write_files'''
        fasta_file = os.path.join(data_dir, 'megares_zip_parser_write_files', 'megares_database_v1.01.fasta')
        annotations_file = os.path.join(data_dir, 'megares_zip_parser_write_files', 'megares_annotations_v1.01.csv')
        mappings_file = os.path.join(data_dir, 'megares_zip_parser_write_files', 'megares_to_external_header_mappings_v1.01.tsv')
        sequences = {}
        pyfastaq.tasks.file_to_dict(fasta_file, sequences)
        annotation_data = megares_zip_parser.MegaresZipParser._load_annotations_file(annotations_file)
        mappings_data = megares_zip_parser.MegaresZipParser._load_header_mappings_file(mappings_file)

        tmp_prefix = 'tmp.test_megares_zip_parser_write_files'
        megares_zip_parser.MegaresZipParser._write_files(tmp_prefix, sequences, annotation_data, mappings_data)

        expected_fasta = os.path.join(data_dir, 'megares_zip_parser_write_files.expect.fa')
        expected_tsv = os.path.join(data_dir, 'megares_zip_parser_write_files.expect.tsv')
        self.assertTrue(filecmp.cmp(expected_fasta, tmp_prefix + '.fa', shallow=False))
        self.assertTrue(filecmp.cmp(expected_tsv, tmp_prefix + '.tsv', shallow=False))
        os.unlink(tmp_prefix + '.fa')
        os.unlink(tmp_prefix + '.tsv')



import unittest
import os
import filecmp
from ariba import flag, report_filter, report

modules_dir = os.path.dirname(os.path.abspath(report_filter.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestReportFilter(unittest.TestCase):
    def test_init_good_file(self):
        '''test __init__ on good input file'''
        infile = os.path.join(data_dir, 'report_filter_test_init_good.tsv')
        rf = report_filter.ReportFilter(infile=infile)
        line1 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.1', '1300', '10.5', '1', 'SNP', 'n', 'C42T', '0', '.', '.', '42', '42', 'C', '142', '142', 'C', '500', 'C', '500', 'a:n:C42T:id1:foo', 'free_text'])
        line2 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.1', '1300', '10.5', '1', 'SNP', 'n', 'A51G', '0', '.', '.', '51', '51', 'C', '151', '151', 'C', '542', 'C', '542', 'a:n:A51G:id2:bar', 'free_text2'])
        line3 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.2', '1300', '12.4', '1', 'SNP', 'n', 'A51G', '0', '.', '.', '51', '51', 'C', '151', '151', 'C', '542', 'C', '542', 'a:n:A51G:id3:spam', 'free_text3'])
        line4 = '\t'.join(['ariba_cluster2', 'cluster2', '1', '0', '179', '20000', 'cluster2', '1042', '1042', '42.42', 'cluster2.scaffold.1', '1442', '20.2', '1', 'SNP', 'p', 'I42L', '1', 'I42L', 'NONSYN', '112', '112', 'C', '442', '442', 'T', '300', 'T', '290', 'a:v:I42L:id4:eggs', 'free_text3'])

        expected = {
            'cluster1': {
                'cluster1.scaffold.1': [report_filter.ReportFilter._report_line_to_dict(line1), report_filter.ReportFilter._report_line_to_dict(line2)],
                'cluster1.scaffold.2': [report_filter.ReportFilter._report_line_to_dict(line3)],
            },
            'cluster2': {
                'cluster2.scaffold.1': [report_filter.ReportFilter._report_line_to_dict(line4)]
            }
        }

        self.assertEqual(expected, rf.report)


    def test_init_bad_file(self):
        '''test __init__ on bad input file'''
        infile = os.path.join(data_dir, 'report_filter_test_init_bad.tsv')
        with self.assertRaises(report_filter.Error):
            report_filter.ReportFilter(infile=infile)


    def test_report_line_to_dict(self):
        line = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t99.42\tcluster1.scaffold.1\t999\t23.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\ta:n:C42T:id1:foo\tfree text'
        expected = {
            'ariba_ref_name':     'ariba_cluster1',
            'ref_name':           'cluster1',
            'gene':               '0',
            'var_only':           '0',
            'flag':               flag.Flag(27),
            'reads':              10000,
            'cluster':        'cluster1',
            'ref_len':            1000,
            'ref_base_assembled': 999,
            'pc_ident':           99.42,
            'ctg':                'cluster1.scaffold.1',
            'ctg_len':            999,
            'ctg_cov':            23.2,
            'known_var':          '1',
            'var_type':           'SNP',
            'var_seq_type':       'n',
            'known_var_change':   'C42T',
            'has_known_var':      '0',
            'ref_ctg_change':     '.',
            'ref_ctg_effect':     '.',
            'ref_start':          42,
            'ref_end':            42,
            'ref_nt':             'C',
            'ctg_start':          142,
            'ctg_end':            142,
            'ctg_nt':             'C',
            'smtls_total_depth':  '500',
            'smtls_nts':       '.',
            'smtls_nts_depth':    '500',
            'var_description':    'a:n:C42T:id1:foo',
            'free_text':          'free text',
        }

        self.assertEqual(expected, report_filter.ReportFilter._report_line_to_dict(line))

        bad_line = '\t'.join(line.split('\t')[:-1])
        self.assertEqual(None, report_filter.ReportFilter._report_line_to_dict(bad_line))


    def test_dict_to_report_line(self):
        '''Test _dict_to_report_line'''
        report_dict = {
            'ariba_ref_name':     'ariba_cluster1',
            'ref_name':           'cluster1',
            'gene':               '0',
            'var_only':           '0',
            'flag':               flag.Flag(27),
            'reads':              10000,
            'cluster':        'cluster1',
            'ref_len':            1000,
            'ref_base_assembled': 999,
            'pc_ident':           99.42,
            'ctg':                'cluster1.scaffold.1',
            'ctg_len':            1300,
            'ctg_cov':            42.4,
            'known_var':          '1',
            'var_type':           'SNP',
            'var_seq_type':       'n',
            'known_var_change':   'C42T',
            'has_known_var':      '0',
            'ref_ctg_change':     '.',
            'ref_ctg_effect':     '.',
            'ref_start':          42,
            'ref_end':            42,
            'ref_nt':             'C',
            'ctg_start':          142,
            'ctg_end':            142,
            'ctg_nt':             'C',
            'smtls_total_depth':  '500',
            'smtls_nts':       '.',
            'smtls_nts_depth':    '500',
            'var_description':    'a:n:C42T:id1:foo',
            'free_text':          'free text',
        }

        expected = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t99.42\tcluster1.scaffold.1\t1300\t42.4\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\ta:n:C42T:id1:foo\tfree text'
        self.assertEqual(expected, report_filter.ReportFilter._dict_to_report_line(report_dict))


    def test_load_report(self):
        good_infile = os.path.join(data_dir, 'report_filter_test_load_report_good.tsv')
        bad_infile = os.path.join(data_dir, 'report_filter_test_load_report_bad.tsv')

        line1 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.1', '1300', '12.2', '1', 'SNP', 'n', 'C42T', '0', '.', '.', '42', '42', 'C', '142', '142', 'C', '500', 'C', '500', 'a:n:C42T:id1:foo', 'free_text'])
        line2 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.1', '1300', '12.2', '1', 'SNP', 'n', 'A51G', '0', '.', '.', '51', '51', 'C', '151', '151', 'C', '542', 'C', '542', 'a:n:A51G:id2:bar', 'free_text2'])
        line3 = '\t'.join(['ariba_cluster1', 'cluster1', '0', '0', '27', '10000', 'cluster1', '1000', '999', '99.42', 'cluster1.scaffold.2', '1300', '22.2', '1', 'SNP', 'n', 'A51G', '0', '.', '.', '51', '51', 'C', '151', '151', 'C', '542', 'C', '542', 'a:n:A51G:id3:spam', 'free_text3'])
        line4 = '\t'.join(['ariba_cluster2', 'cluster2', '1', '1', '179', '20000', 'cluster2', '1042', '1042', '42.42', 'cluster2.scaffold.1', '1442', '33.3', '1', 'SNP', 'p', 'I42L', '1', 'I42L', 'NONSYN', '112', '112', 'C', '442', '442', 'T', '300', 'T', '290', 'a:v:I42L:id4:eggs', 'free_text3'])

        expected = {
            'cluster1': {
                'cluster1.scaffold.1': [report_filter.ReportFilter._report_line_to_dict(line1), report_filter.ReportFilter._report_line_to_dict(line2)],
                'cluster1.scaffold.2': [report_filter.ReportFilter._report_line_to_dict(line3)],
            },
            'cluster2': {
                'cluster2.scaffold.1': [report_filter.ReportFilter._report_line_to_dict(line4)]
            }
        }

        got = report_filter.ReportFilter._load_report(good_infile)
        self.maxDiff = None
        self.assertEqual(expected, got)
        with self.assertRaises(report_filter.Error):
            report_filter.ReportFilter._load_report(bad_infile)


    def test_report_dict_passes_non_essential_filters_known_vars(self):
        '''Test _report_dict_passes_non_essential_filters with known vars'''
        tests = [
            ('.', '.', True, True),
            ('.', '.', False, True),
            ('0', '0', True, True),
            ('0', '0', False, True),
            ('1', '0', True, False),
            ('1', '1', True, True),
            ('1', '0', False, True),
            ('1', '1', False, True),
        ]

        for known_var, has_known_var, ignore_not_has_known_variant, expected in tests:
            d = {'known_var': known_var, 'has_known_var': has_known_var}
            rf = report_filter.ReportFilter(ignore_not_has_known_variant=ignore_not_has_known_variant)
            self.assertEqual(expected, rf._report_dict_passes_non_essential_filters(d))


    def test_report_dict_passes_non_essential_filters_synonymous(self):
        '''Test _report_dict_passes_non_essential_filters with synonymous AA changes'''
        tests = [
             ('.', True, True),
             ('.', False, True),
             ('SNP', True, True),
             ('SNP', False, True),
             ('SYN', True, False),
             ('SYN', False, True),
        ]

        for var, remove_synonymous_snps, expected in tests:
            d = {'known_var': '1', 'ref_ctg_effect': var, 'has_known_var': '1'}
            rf = report_filter.ReportFilter(remove_synonymous_snps=remove_synonymous_snps)
            self.assertEqual(expected, rf._report_dict_passes_non_essential_filters(d))


    def test_report_dict_passes_essential_filters(self):
        '''Test _report_dict_passes_essential_filters'''
        line1 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t98.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        line2 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t0\t98.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        line3 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t78.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        tests = [
            (report_filter.ReportFilter._report_line_to_dict(line1), True),
            (report_filter.ReportFilter._report_line_to_dict(line2), False),
            (report_filter.ReportFilter._report_line_to_dict(line3), False),
        ]

        for test_dict, expected in tests:
            rf = report_filter.ReportFilter()
            self.assertEqual(expected,  rf._report_dict_passes_essential_filters(test_dict))


    def test_flag_passes_filter(self):
        '''Test _flag_passes_filter'''
        rf = report_filter.ReportFilter()
        exclude_flags = ['assembly_fail', 'ref_seq_choose_fail']
        f = flag.Flag()
        self.assertTrue(rf._flag_passes_filter(f, exclude_flags))
        f.add('assembled')
        self.assertTrue(rf._flag_passes_filter(f, exclude_flags))
        f = flag.Flag()
        f.add('assembly_fail')
        self.assertFalse(rf._flag_passes_filter(f, exclude_flags))
        f = flag.Flag()
        f.add('ref_seq_choose_fail')
        self.assertFalse(rf._flag_passes_filter(f, exclude_flags))


    def test_filter_list_of_dicts_all_fail(self):
        '''Test _filter_list_of_dicts where all fail'''
        rf = report_filter.ReportFilter()
        line1 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t88.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        line2 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t78.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        dict1 = report_filter.ReportFilter._report_line_to_dict(line1)
        dict2 = report_filter.ReportFilter._report_line_to_dict(line2)
        got = rf._filter_list_of_dicts([dict1, dict2])
        self.assertEqual([], got)


    def test_filter_list_of_dicts_with_essential(self):
        '''Test _filter_list_of_dicts with an essential line but all others fail'''
        rf = report_filter.ReportFilter(ignore_not_has_known_variant=True)
        line1 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t98.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        line2 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t78.42\tcluster1.scaffold.1\t400\t12.2\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        dict1 = report_filter.ReportFilter._report_line_to_dict(line1)
        dict2 = report_filter.ReportFilter._report_line_to_dict(line2)
        expected_line = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t98.42\tcluster1.scaffold.1\t400\t12.2\t' + '\t'.join(['.'] * 17) + '\tfree text'
        expected = [report_filter.ReportFilter._report_line_to_dict(expected_line)]
        assert expected != [None]
        got = rf._filter_list_of_dicts([dict1, dict2])
        self.assertEqual(expected, got)


    def test_filter_list_of_dicts_with_pass(self):
        '''Test _filter_list_of_dicts with a line that passes'''
        rf = report_filter.ReportFilter(ignore_not_has_known_variant=True)
        line1 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t98.42\tcluster1.scaffold.1\t500\t12.1\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        line2 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t98.42\tcluster1.scaffold.1\t500\t12.1\t1\tSNP\tn\tC46T\t1\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C46T\tfree text'
        line3 = 'ariba_cluster1\tcluster1\t0\t0\t27\t10000\tcluster1\t1000\t999\t78.42\tcluster1.scaffold.1\t500\t12.1\t1\tSNP\tn\tC42T\t0\t.\t.\t42\t42\tC\t142\t142\tC\t500\t.\t500\tDescription_of_variant C42T\tfree text'
        dict1 = report_filter.ReportFilter._report_line_to_dict(line1)
        dict2 = report_filter.ReportFilter._report_line_to_dict(line2)
        dict3 = report_filter.ReportFilter._report_line_to_dict(line3)
        got = rf._filter_list_of_dicts([dict1, dict2, dict3])
        self.assertEqual([dict2], got)


    def test_remove_all_after_first_frameshift(self):
        '''Test _remove_all_after_first_frameshift'''
        self.assertEqual([], report_filter.ReportFilter._remove_all_after_first_frameshift([]))
        line1 = 'ariba_cluster1\tcluster1\t1\t0\t528\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tE89G\tNONSYN\t65\t265\tA;A\t766\t766\tG;C\t88;90\t.;.\t87;90\t.\t.'
        line2 = 'ariba_cluster1\tcluster1\t1\t0\t528\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tQ37fs\tFSHIFT\t109\t109\tA\t634\t634\t.\t67\t.\t67\t.\t.'
        line3 = 'ariba_cluster1\tcluster1\t1\t0\t528\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tE89G\tNONSYN\t265\t265\tA;A\t766\t766\tG;C\t88;90\t.;.\t87;90\t.\t.'
        dict1 = report_filter.ReportFilter._report_line_to_dict(line1)
        dict2 = report_filter.ReportFilter._report_line_to_dict(line2)
        dict3 = report_filter.ReportFilter._report_line_to_dict(line3)
        self.assertEqual([dict1], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1]))
        self.assertEqual([dict1, dict2], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1, dict2]))
        self.assertEqual([dict2], report_filter.ReportFilter._remove_all_after_first_frameshift([dict2, dict3]))
        self.assertEqual([dict1, dict2], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1, dict2, dict3]))


    def test_remove_all_after_first_frameshift_complete_gene(self):
        '''Test _remove_all_after_first_frameshift when gene is complete'''
        self.assertEqual([], report_filter.ReportFilter._remove_all_after_first_frameshift([]))
        line1 = 'ariba_cluster1\tcluster1\t1\t0\t538\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tE89G\tNONSYN\t65\t265\tA;A\t766\t766\tG;C\t88;90\t.;.\t87;90\t.\t.'
        line2 = 'ariba_cluster1\tcluster1\t1\t0\t538\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tQ37fs\tFSHIFT\t109\t109\tA\t634\t634\t.\t67\t.\t67\t.\t.'
        line3 = 'ariba_cluster1\tcluster1\t1\t0\t538\t1874\tcluster1\t1188\t1097\t92.43\tcluster1.scaffold.1\t2218\t42.42\t0\t.\tp\t.\t0\tE89G\tNONSYN\t265\t265\tA;A\t766\t766\tG;C\t88;90\t.;.\t87;90\t.\t.'
        dict1 = report_filter.ReportFilter._report_line_to_dict(line1)
        dict2 = report_filter.ReportFilter._report_line_to_dict(line2)
        dict3 = report_filter.ReportFilter._report_line_to_dict(line3)
        self.assertEqual([dict1], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1]))
        self.assertEqual([dict1, dict2], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1, dict2]))
        self.assertEqual([dict2, dict3], report_filter.ReportFilter._remove_all_after_first_frameshift([dict2, dict3]))
        self.assertEqual([dict1, dict2, dict3], report_filter.ReportFilter._remove_all_after_first_frameshift([dict1, dict2, dict3]))


    def test_filter_dicts(self):
        '''Test _filter_dicts'''
        rf = report_filter.ReportFilter(min_ref_base_assembled=10, ignore_not_has_known_variant=True)
        ref_2_dict = {x: '.' for x in report.columns}
        ref_2_dict['pc_ident'] = 91.0
        ref_2_dict['ref_base_assembled'] = 10
        ref_2_dict['has_known_var'] = '0'
        ref_2_dict['flag'] = flag.Flag(27)
        ref_2_dict['var_type'] = '.'

        rf.report = {
            'ref1': {
                'ref1.scaff1': [
                    {'flag': flag.Flag(27), 'pc_ident': 91.0, 'ref_base_assembled': 9, 'known_var': '1', 'has_known_var': '1', 'var_type': 'SNP'},
                    {'flag': flag.Flag(27), 'pc_ident': 91.5, 'ref_base_assembled': 11, 'known_var': '1', 'has_known_var': '1', 'var_type': 'HET'},
                    {'flag': flag.Flag(27), 'pc_ident': 89.0, 'ref_base_assembled': 10, 'known_var': '1', 'has_known_var': '1', 'var_type': 'SNP'},
                    {'flag': flag.Flag(27), 'pc_ident': 90.0, 'ref_base_assembled': 11, 'known_var': '1', 'has_known_var': '0', 'var_type': 'SNP'},
                    {'flag': flag.Flag(27), 'pc_ident': 90.0, 'ref_base_assembled': 11, 'known_var': '1', 'has_known_var': '1', 'var_type': 'SNP'},
                ]
            },
            'ref2': {
                'ref2.scaff1': [
                    ref_2_dict
                ]
            },
            'ref3': {
                'ref3.scaff1': [
                    {'flag': flag.Flag(27), 'pc_ident': 84.0, 'ref_base_assembled': 10, 'known_var': '1', 'has_known_var': '0', 'var_type': 'SNP'},
                ]
            },
            'ref4': {
                'ref4.scaff1': [
                    {'flag': flag.Flag(64), 'pc_ident': '.', 'ref_base_assembled': '.', 'known_var': '.', 'has_known_var': '.', 'var_type': '.'},
                ]
            }
        }

        expected = {
            'ref1': {
                'ref1.scaff1': [
                    {'flag': flag.Flag(27), 'pc_ident': 91.5, 'ref_base_assembled': 11, 'known_var': '1', 'has_known_var': '1', 'var_type': 'HET'},
                    {'flag': flag.Flag(27), 'pc_ident': 90.0, 'ref_base_assembled': 11, 'known_var': '1', 'has_known_var': '1', 'var_type': 'SNP'},
                ]
            },
            'ref2': {
                'ref2.scaff1': [ref_2_dict]
            }
        }

        rf._filter_dicts()
        self.assertEqual(expected, rf.report)


    def test_write_report_tsv(self):
        '''Test write_report_tsv'''
        infile = os.path.join(data_dir, 'report_filter_test_write_report.tsv')
        tmpfile = 'tmp.test.report_filter.write_report.tsv'
        rf = report_filter.ReportFilter(infile=infile)
        rf._write_report_tsv(tmpfile)
        self.assertTrue(filecmp.cmp(tmpfile, infile, shallow=False))
        os.unlink(tmpfile)


    def test_run(self):
        '''Test run'''
        infile = os.path.join(data_dir, 'report_filter_test_run.in.tsv')
        expected_file = os.path.join(data_dir, 'report_filter_test_run.expected.tsv')
        tmpfile = 'tmp.test.report_filter.run.out.tsv'
        rf = report_filter.ReportFilter(infile=infile)
        rf.run(tmpfile)
        self.assertTrue(filecmp.cmp(expected_file, tmpfile, shallow=False))
        os.unlink(tmpfile)




import pyfastaq

from ariba import flag

class Error (Exception): pass

class ReportFlagExpander:
    def __init__(self, infile, outfile):
        self.infile = infile
        self.outfile = outfile


    def run(self):
        f_in = pyfastaq.utils.open_file_read(self.infile)
        f_out = pyfastaq.utils.open_file_write(self.outfile)
        flag_index = None

        for line in f_in:
            fields = line.rstrip().split()

            if flag_index is None:
                try:
                    flag_index = fields.index('flag')
                except:
                    raise Error('"flag" column not found in first line of file ' + self.infile +'. Cannot continue')
            else:
                f = flag.Flag(int(fields[flag_index]))
                fields[flag_index] = f.to_comma_separated_string()

            print(*fields, sep='\t', file=f_out)

        f_in.close()
        f_out.close()




import unittest
import os
from ariba import megares_data_finder

modules_dir = os.path.dirname(os.path.abspath(megares_data_finder.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestMegaresDataFinder(unittest.TestCase):
    def test_zips_from_index_page_string(self):
        '''test _zips_from_index_page_string'''
        html_string = r''''<!doctype html>
<html>
  <head>

  </head>

<ul>
  <li><a href="megares_v1.01.zip">All Files</a></li>
  <li><a href="foo.zip">All Files</a></li>
  <li><a href="megares_v1.00.zip">All Files</a></li>
</html>'''

        expected = {'1.00': 'megares_v1.00.zip', '1.01': 'megares_v1.01.zip'}
        got = megares_data_finder.MegaresDataFinder._zips_from_index_page_string(html_string)
        self.assertEqual(expected, got)


    def test_get_url_for_version(self):
        '''test _get_url_for_version'''
        zips = {'1.00': 'megares_v1.00.zip', '1.01': 'megares_v1.01.zip'}
        self.assertEqual('megares_v1.01.zip', megares_data_finder.MegaresDataFinder._get_url_for_version(zips))
        self.assertEqual('megares_v1.00.zip', megares_data_finder.MegaresDataFinder._get_url_for_version(zips, version='1.00'))
        with self.assertRaises(megares_data_finder.Error):
            self.assertEqual('megares_v1.00.zip', megares_data_finder.MegaresDataFinder._get_url_for_version(zips, version='0.42'))




import unittest
import os
from ariba import summary_cluster, summary_sample

modules_dir = os.path.dirname(os.path.abspath(summary_sample.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestSummarySample(unittest.TestCase):
    def test_load_file(self):
        '''Test _load_file'''
        infile = os.path.join(data_dir, 'summary_sample_test_load_file.in.tsv')
        with open(infile) as f:
             lines = [x.rstrip() for x in f]

        dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines[1:]]
        cluster1 = summary_cluster.SummaryCluster()
        cluster1.add_data_dict(dicts[0])
        cluster1.add_data_dict(dicts[1])
        cluster1.add_data_dict(dicts[2])
        cluster1.gather_data()
        cluster2 = summary_cluster.SummaryCluster()
        cluster2.add_data_dict(dicts[3])
        cluster2.add_data_dict(dicts[4])
        cluster2.gather_data()
        cluster3 = summary_cluster.SummaryCluster()
        cluster3.add_data_dict(dicts[5])
        cluster3.gather_data()

        expected = {
            'cluster.n': cluster1,
            'cluster.p': cluster2,
            'cluster.v': cluster3
        }

        got = summary_sample.SummarySample._load_file(infile, 90)
        self.assertEqual(expected, got)

        got = summary_sample.SummarySample._load_file(infile, 90, only_clusters={'cluster.n'})
        expected = {'cluster.n': cluster1}
        self.assertEqual(expected, got)

    def test_column_summary_data(self):
        '''Test _column_summary_data'''
        infile = os.path.join(data_dir, 'summary_sample_test_column_summary_data.tsv')
        sample_summary = summary_sample.SummarySample(infile)
        sample_summary.clusters = sample_summary._load_file(infile, 90)
        expected = {
            'cluster.n': {
                'assembled': 'yes',
                'match': 'yes',
                'ref_seq': 'noncoding1',
                'known_var': 'yes',
                'novel_var': 'yes',
                'pct_id': '98.33',
                'ctg_cov': '35.4',
            },
            'cluster.p': {
                'assembled': 'yes',
                'match': 'yes',
                'ref_seq': 'presence_absence1',
                'known_var': 'yes',
                'novel_var': 'no',
                'pct_id': '98.96',
                'ctg_cov': '35.1',
            },
            'cluster.v': {
                'assembled': 'yes',
                'match': 'yes',
                'ref_seq': 'variants_only1',
                'known_var': 'yes',
                'novel_var': 'no',
                'pct_id': '100.0',
                'ctg_cov': '42.4',
            }
        }
        self.maxDiff = None
        got = sample_summary._column_summary_data()
        self.assertEqual(expected, got)


    def test_var_groups(self):
        '''test _var_groups'''
        infile = os.path.join(data_dir, 'summary_sample_test_var_groups.tsv')
        sample_summary = summary_sample.SummarySample(infile)
        sample_summary.clusters = sample_summary._load_file(infile, 90)
        got = sample_summary._var_groups()
        expected = {
            'cluster.n': {'id1', 'id2'},
            'cluster.p': {'id3'},
            'cluster.v': {'id4'}
        }
        self.assertEqual(expected, got)


    def test_variant_column_names_tuples_and_het_snps(self):
        '''Test _variant_column_names_tuples_and_het_snps'''
        infile = os.path.join(data_dir, 'summary_sample_test_column_names_tuples_and_het_snps.tsv')
        sample_summary = summary_sample.SummarySample(infile)
        sample_summary.clusters = sample_summary._load_file(infile, 90)
        sample_summary.column_summary_data = sample_summary._column_summary_data()
        expected_column_names = {
            'cluster.v': {('variants_only1', 'S5T', 'ungrouped', None)},
            'cluster.n': {
                ('noncoding1', 'A6G', 'grouped', 'id2'),
                ('noncoding1', 'A14T', 'ungrouped', None),
                ('noncoding1', 'G15T', 'novel', None)
             },
            'cluster.p': {('presence_absence1', 'A10V', 'grouped', 'id3')}
        }
        got_column_names, got_het_snps = sample_summary._variant_column_names_tuples_and_het_snps()
        self.assertEqual(expected_column_names, got_column_names)

        expected_het_snps = {
            'cluster.v': {},
            'cluster.n': {'.': {'A14T': 80.0}},
            'cluster.p': {},
        }
        self.assertEqual(expected_het_snps, got_het_snps)




import unittest
import os
import filecmp
from ariba import common, external_progs, ref_preparer

modules_dir = os.path.dirname(os.path.abspath(ref_preparer.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestRefPreparer(unittest.TestCase):
    def test_fasta_to_metadata(self):
        '''test _fasta_to_metadata'''
        infile = os.path.join(data_dir, 'ref_preparer_test_fasta_to_metadata.fa')
        tmp_out = 'tmp.test_fasta_to_metadata.tsv'
        expected_coding = os.path.join(data_dir, 'ref_preparer_test_fasta_to_metadata.coding.tsv')
        expected_noncoding = os.path.join(data_dir, 'ref_preparer_test_fasta_to_metadata.noncoding.tsv')

        with open(tmp_out, 'w') as f:
            ref_preparer.RefPreparer._fasta_to_metadata(infile, f, True)
        self.assertTrue(filecmp.cmp(expected_coding, tmp_out, shallow=False))

        with open(tmp_out, 'w') as f:
            ref_preparer.RefPreparer._fasta_to_metadata(infile, f, False)
        self.assertTrue(filecmp.cmp(expected_noncoding, tmp_out, shallow=False))

        os.unlink(tmp_out)


    def test_rename_clusters(self):
        '''test _rename_clusters'''
        clusters_in = {
           '0': {'no_dot_in_name'},
           '1': {'another_no_dot_in_name'},
           '2': {'foo.blah_blah_blah', 'foo.xyz'},
           '3': {'foo.abc', 'foo.def'},
           '4': {'pre1.abc', 'pre2.abc'},
           '5': {'pre1.def', 'pre2.pqr', 'pre2.zxy'},
           '6': {'prefix1.abc', 'prefix1.def', 'something_else.abc'},
           '7': {'prefix1.fgh', 'prefix1.ijk', 'something_else_again.abc'},
           '8': {'xyz.1', 'xyz.2', 'abcdefgh'},
           '9': {'a.foo', 'a.bar'},
           '10': {'abc_1.1'},
           '11': {'abc.2'},
           '12': {'abc.3'},
           '13': {'abc.4'},
           '14': {'def.1'},
           '15': {'def_1.2'},
           '16': {'def_2.3'},
           '17': {'def.4'},
           '18': {'def.5'},
           '19': {'x_1.foo'},
           '20': {'x_1.bar'},
           '21': {'x_1.baz'},
           '22': {'x_1_2.abc'},
           '23': {'x_1_2.def'},
           '24': {'y_1.foo'},
           '25': {'y_1_2.def'},
           '26': {'y_1.bar'},
           '27': {'y_1.baz'},
           '28': {'y_1_2.abc'},
        }

        expected = {
            'cluster': {'no_dot_in_name'},
            'cluster_1': {'another_no_dot_in_name'},
            'foo': {'foo.blah_blah_blah', 'foo.xyz'},
            'foo_1': {'foo.abc', 'foo.def'},
            'pre-': {'pre1.abc', 'pre2.abc'},
            'pre-_1': {'pre1.def', 'pre2.pqr', 'pre2.zxy'},
            'prefix1+': {'prefix1.abc', 'prefix1.def', 'something_else.abc'},
            'prefix1+_1': {'prefix1.fgh', 'prefix1.ijk', 'something_else_again.abc'},
            'xyz+': {'xyz.1', 'xyz.2', 'abcdefgh'},
            'cluster_2': {'a.foo', 'a.bar'},
            'abc_1': {'abc_1.1'},
            'abc': {'abc.2'},
            'abc_2': {'abc.3'},
            'abc_3': {'abc.4'},
            'def_1': {'def_1.2'},
            'def_2': {'def_2.3'},
            'def': {'def.1'},
            'def_3': {'def.4'},
            'def_4': {'def.5'},
            'x_1': {'x_1.foo'},
            'x_1_1': {'x_1.bar'},
            'x_1_2_1': {'x_1_2.abc'},
            'x_1_2_2': {'x_1_2.def'},
            'x_1_2': {'x_1.baz'},
            'y_1': {'y_1.foo'},
            'y_1_1': {'y_1.bar'},
            'y_1_2_1': {'y_1_2.abc'},
            'y_1_2': {'y_1_2.def'},
            'y_1_3': {'y_1.baz'},
        }

        got = ref_preparer.RefPreparer._rename_clusters(clusters_in)
        self.assertEqual(expected, got)


    def test_run(self):
        '''test run'''
        fasta_in = [
            os.path.join(data_dir, 'ref_preparer_test_run.in.1.fa'),
            os.path.join(data_dir, 'ref_preparer_test_run.in.2.fa'),
            os.path.join(data_dir, 'ref_preparer_test_run.in.3.fa'),
        ]
        tsv_in = [
            os.path.join(data_dir, 'ref_preparer_test_run.in.1.tsv'),
            os.path.join(data_dir, 'ref_preparer_test_run.in.2.tsv'),
        ]

        extern_progs = external_progs.ExternalProgs()
        refprep = ref_preparer.RefPreparer(fasta_in, extern_progs, metadata_tsv_files=tsv_in, genetic_code=1)
        tmp_out = 'tmp.ref_preparer_test_run'
        refprep.run(tmp_out)
        expected_outdir = os.path.join(data_dir, 'ref_preparer_test_run.out')

        test_files = [
            '01.filter.check_metadata.tsv',
            '01.filter.check_genes.log',
            '01.filter.check_noncoding.log',
            '01.filter.check_metadata.log',
            '02.cdhit.all.fa',
            '02.cdhit.clusters.tsv',
            '02.cdhit.gene.fa',
            '02.cdhit.gene.varonly.fa',
            '02.cdhit.noncoding.fa',
            '02.cdhit.noncoding.varonly.fa',
        ]

        for filename in test_files:
            expected = os.path.join(expected_outdir, filename)
            got = os.path.join(tmp_out, filename)
            self.assertTrue(filecmp.cmp(expected, got, shallow=False))

        common.rmtree(tmp_out)


    def test_run_all_noncoding(self):
        '''test run with no metadata input, all sequences are noncoding'''
        fasta_in = [
            os.path.join(data_dir, 'ref_preparer_test_run.in.1.fa'),
            os.path.join(data_dir, 'ref_preparer_test_run.in.2.fa'),
            os.path.join(data_dir, 'ref_preparer_test_run.in.3.fa'),
        ]

        extern_progs = external_progs.ExternalProgs()
        refprep = ref_preparer.RefPreparer(fasta_in, extern_progs, all_coding='no', genetic_code=1)
        tmp_out = 'tmp.ref_preparer_test_run'
        refprep.run(tmp_out)
        expected_outdir = os.path.join(data_dir, 'ref_preparer_test_run_all_noncoding.out')

        test_files = [
            '00.auto_metadata.tsv',
            '01.filter.check_metadata.tsv',
            '01.filter.check_genes.log',
            '01.filter.check_noncoding.log',
            '01.filter.check_metadata.log',
            '02.cdhit.all.fa',
            '02.cdhit.clusters.tsv',
            '02.cdhit.gene.fa',
            '02.cdhit.gene.varonly.fa',
            '02.cdhit.noncoding.fa',
            '02.cdhit.noncoding.varonly.fa',
        ]

        for filename in test_files:
            expected = os.path.join(expected_outdir, filename)
            got = os.path.join(tmp_out, filename)
            self.assertTrue(filecmp.cmp(expected, got, shallow=False))

        common.rmtree(tmp_out)

    def test_run_noncoding_checks(self):
        '''test run with noncoding sequences that are outside of the allowed size range'''
        fasta_in = [
            os.path.join(data_dir, 'ref_preparer_test_run.in.4.fa')
        ]
        tsv_in = [
            os.path.join(data_dir, 'ref_preparer_test_run.in.4.tsv')
        ]

        extern_progs = external_progs.ExternalProgs()
        refprep = ref_preparer.RefPreparer(
            fasta_in, extern_progs, min_noncoding_length=6, max_noncoding_length=20,
            metadata_tsv_files=tsv_in, genetic_code=1)
        tmp_out = 'tmp.ref_preparer_test_run_noncoding_checks'
        refprep.run(tmp_out)
        expected_outdir = os.path.join(data_dir, 'ref_preparer_test_run_noncoding_checks.out')

        test_files = [
            '01.filter.check_metadata.tsv',
            '01.filter.check_genes.log',
            '01.filter.check_noncoding.log',
            '01.filter.check_metadata.log',
            '02.cdhit.all.fa',
            '02.cdhit.clusters.tsv',
            '02.cdhit.gene.fa',
            '02.cdhit.gene.varonly.fa',
            '02.cdhit.noncoding.fa',
            '02.cdhit.noncoding.varonly.fa',
        ]

        for filename in test_files:
            expected = os.path.join(expected_outdir, filename)
            got = os.path.join(tmp_out, filename)
            self.assertTrue(filecmp.cmp(expected, got, shallow=False))

        common.rmtree(tmp_out)





import unittest
import os
import filecmp
from ariba import common

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestCommon(unittest.TestCase):
    def test_cat_files(self):
        '''test cat_files'''
        infiles = [
            os.path.join(data_dir, 'test_common_cat_files.in.1'),
            os.path.join(data_dir, 'test_common_cat_files.in.2'),
            os.path.join(data_dir, 'test_common_cat_files.in.3'),
        ]
        tmp_out = 'tmp.test.common_cat_files.out'
        expected = os.path.join(data_dir, 'test_common_cat_files.out')
        common.cat_files(infiles, tmp_out)
        self.assertTrue(filecmp.cmp(expected, tmp_out, shallow=False))
        os.unlink(tmp_out)


    def test_rmtree(self):
        '''test rmtree'''
        tmp_dir = 'tmp.rmtree'
        os.mkdir(tmp_dir)
        with open (os.path.join(tmp_dir, 'foo'), 'w') as f:
            pass

        self.assertTrue(os.path.exists(tmp_dir))
        common.rmtree(tmp_dir)
        self.assertFalse(os.path.exists(tmp_dir))





import argparse
from ariba import pubmlst_ref_preparer


def run(options):
    preparer = pubmlst_ref_preparer.PubmlstRefPreparer(
        options.species,
        options.outdir,
        verbose=options.verbose
    )
    preparer.run()




import unittest
import os
import filecmp
import pyfastaq
from ariba import assembly, common
from ariba import external_progs

modules_dir = os.path.dirname(os.path.abspath(assembly.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
extern_progs = external_progs.ExternalProgs(using_spades=True)

class TestAssembly(unittest.TestCase):
    def test_run_fermilite(self):
        '''test _run_fermilite'''
        reads = os.path.join(data_dir, 'assembly_run_fermilite.reads.fq')
        tmp_fa = 'tmp.test_run_fermilite.fa'
        tmp_log = 'tmp.test_run_fermilite.log'
        expected_fa = os.path.join(data_dir, 'assembly_run_fermilite.expected.fa')
        expected_log = os.path.join(data_dir, 'assembly_run_fermilite.expected.log')
        got = assembly.Assembly._run_fermilite(reads, tmp_fa, tmp_log, 'contig')
        self.assertEqual(0, got)
        self.assertTrue(filecmp.cmp(expected_fa, tmp_fa, shallow=False))
        self.assertTrue(filecmp.cmp(expected_log, tmp_log, shallow=False))
        os.unlink(tmp_fa)
        os.unlink(tmp_log)


    def test_run_fermilite_fails(self):
        '''test _run_fermilite when it fails'''
        reads = os.path.join(data_dir, 'assembly_run_fermilite_fail.reads.fq')
        tmp_fa = 'tmp.test_run_fermilite_fails.fa'
        tmp_log = 'tmp.test_run_fermilite_fails.log'
        expected_log = os.path.join(data_dir, 'assembly_run_fermilite_fails.expected.log')
        got = assembly.Assembly._run_fermilite(reads, tmp_fa, tmp_log, 'contig')
        self.assertEqual(1, got)
        self.assertFalse(os.path.exists(tmp_fa))
        self.assertTrue(filecmp.cmp(expected_log, tmp_log, shallow=False))
        os.unlink(tmp_log)


    def test_assemble_with_fermilite(self):
        '''test _assemble_with_fermilite'''
        reads1 = os.path.join(data_dir, 'assembly_assemble_with_fermilite.reads_1.fq')
        reads2 = os.path.join(data_dir, 'assembly_assemble_with_fermilite.reads_2.fq')
        expected_log = os.path.join(data_dir, 'assembly_assemble_with_fermilite.expected.log')
        expected_fa = os.path.join(data_dir, 'assembly_assemble_with_fermilite.expected.fa')
        tmp_dir = 'tmp.test_assemble_with_fermilite'
        tmp_log = 'tmp.test_assemble_with_fermilite.log'
        tmp_log_fh = open(tmp_log, 'w')
        print('First line', file=tmp_log_fh)
        a = assembly.Assembly(reads1, reads2, 'not needed', 'not needed', tmp_dir, 'not_needed_for_this_test.fa', 'not_needed_for_this_test.bam', tmp_log_fh, 'not needed')
        a._assemble_with_fermilite()
        self.assertTrue(a.assembled_ok)
        tmp_log_fh.close()
        self.assertTrue(filecmp.cmp(expected_log, tmp_log, shallow=False))
        self.assertTrue(filecmp.cmp(expected_fa, os.path.join(tmp_dir, 'debug_all_contigs.fa'), shallow=False))
        common.rmtree(tmp_dir)
        os.unlink(tmp_log)


    def test_assemble_with_fermilite_fails(self):
        '''test _assemble_with_fermilite fails'''
        reads1 = os.path.join(data_dir, 'assembly_assemble_with_fermilite_fails.reads_1.fq')
        reads2 = os.path.join(data_dir, 'assembly_assemble_with_fermilite_fails.reads_2.fq')
        expected_log = os.path.join(data_dir, 'assembly_assemble_with_fermilite_fails.expected.log')
        tmp_dir = 'tmp.test_assemble_with_fermilite_fails'
        tmp_log = 'tmp.test_assemble_with_fermilite_fails.log'
        tmp_log_fh = open(tmp_log, 'w')
        print('First line', file=tmp_log_fh)
        a = assembly.Assembly(reads1, reads2, 'not needed', 'not needed', tmp_dir, 'not_needed_for_this_test.fa', 'not_needed_for_this_test.bam', tmp_log_fh, 'not needed')
        a._assemble_with_fermilite()
        self.assertFalse(a.assembled_ok)
        tmp_log_fh.close()
        self.assertTrue(filecmp.cmp(expected_log, tmp_log, shallow=False))
        self.assertFalse(os.path.exists(os.path.join(tmp_dir, 'contigs.fa')))
        common.rmtree(tmp_dir)
        os.unlink(tmp_log)


    def test_fix_contig_orientation(self):
        '''test _fix_contig_orientation'''
        scaffs_in = os.path.join(data_dir, 'assembly_test_fix_contig_orientation.in.fa')
        expected_out = os.path.join(data_dir, 'assembly_test_fix_contig_orientation.out.fa')
        ref_fa = os.path.join(data_dir, 'assembly_test_fix_contig_orientation.ref.fa')
        tmp_out = 'tmp.assembly_test_fix_contig_orientation.out.fa'
        got = assembly.Assembly._fix_contig_orientation(scaffs_in, ref_fa, tmp_out)
        expected = {'match_both_strands'}
        self.assertTrue(filecmp.cmp(expected_out, tmp_out, shallow=False))
        self.assertEqual(expected, got)
        os.unlink(tmp_out)


    def test_parse_bam(self):
        '''test _parse_bam'''
        bam = os.path.join(data_dir, 'assembly_test_parse_assembly_bam.bam')
        assembly_fa = os.path.join(data_dir, 'assembly_test_parse_assembly_bam.assembly.fa')
        assembly_seqs = {}
        pyfastaq.tasks.file_to_dict(assembly_fa, assembly_seqs)
        self.assertTrue(assembly.Assembly._parse_bam(assembly_seqs, bam, 10, 1000))
        os.unlink(bam + '.soft_clipped')
        os.unlink(bam + '.unmapped_mates')
        os.unlink(bam + '.scaff')

    def test_check_spades_log_file(self):
        '''test _check_spades_log_file'''
        good_file = os.path.join(data_dir, 'assembly_test_check_spades_log_file.log.good')
        bad_file = os.path.join(data_dir, 'assembly_test_check_spades_log_file.log.bad')
        self.assertTrue(assembly.Assembly._check_spades_log_file(good_file))
        with self.assertRaises(assembly.Error):
            self.assertTrue(assembly.Assembly._check_spades_log_file(bad_file))

    @unittest.skipUnless(extern_progs.exe('spades'), "Spades assembler is optional and is not configured")
    def test_assemble_with_spades(self):
        '''test _assemble_with_spades'''
        reads1 = os.path.join(data_dir, 'assembly_test_assemble_with_spades_reads_1.fq')
        reads2 = os.path.join(data_dir, 'assembly_test_assemble_with_spades_reads_2.fq')
        tmp_dir = 'tmp.test_assemble_with_spades'
        tmp_log = 'tmp.test_assemble_with_spades.log'
        with open(tmp_log, 'w') as tmp_log_fh:
            print('First line', file=tmp_log_fh)
            common.rmtree(tmp_dir)
            #using spades_options=" --only-assembler" because error correction cannot determine quality offset on this
            #artificial dataset
            a = assembly.Assembly(reads1, reads2, 'not needed', 'not needed', tmp_dir, 'not_needed_for_this_test.fa',
                                  'not_needed_for_this_test.bam', tmp_log_fh, 'not needed',
                                  assembler="spades", spades_options=" --only-assembler")
            a._assemble_with_spades()
        self.assertTrue(a.assembled_ok)
        common.rmtree(tmp_dir)
        os.unlink(tmp_log)

    @unittest.skipUnless(extern_progs.exe('spades'), "Spades assembler is optional and is not configured")
    def test_assemble_with_spades_fail(self):
        '''test _assemble_with_spades handles spades fail'''
        reads1 = os.path.join(data_dir, 'assembly_test_assemble_with_spades_fails_reads_1.fq')
        reads2 = os.path.join(data_dir, 'assembly_test_assemble_with_spades_fails_reads_2.fq')
        tmp_dir = 'tmp.test_assemble_with_spades_fail'
        tmp_log = 'tmp.test_assemble_with_spades_fail.log'
        with open(tmp_log, 'w') as tmp_log_fh:
            print('First line', file=tmp_log_fh)
            common.rmtree(tmp_dir)
            a = assembly.Assembly(reads1, reads2, 'not needed', 'not needed', tmp_dir, 'not_needed_for_this_test.fa',
                                  'not_needed_for_this_test.bam', tmp_log_fh, 'not needed',
                                  assembler="spades", spades_options=" --only-assembler")
            a._assemble_with_spades()
        self.assertFalse(a.assembled_ok)
        common.rmtree(tmp_dir)
        os.unlink(tmp_log)
